package edu.berkeley.qfat;
import java.io.*;
import java.nio.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

public abstract class Viewer implements GLEventListener, MouseListener, MouseMotionListener, KeyListener, MouseWheelListener  {

    Mesh.Vertex closest = null;

    public Mesh tile = new Mesh(false);
    public Mesh goal = new Mesh(false);

    private Matrix projection = null;

    JFrame f;
    GLCanvas glcanvas;
    Point clickPoint = null;
    Point clickClosest = null;

    int mousex;
    int mousey;
    float tx = 0;
    float ty = 0;
    float tz = 0;
    float anglex = 0;
    float angley = 0;

    boolean alt = false;
    boolean shift = false;
    boolean control = false;

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) { }
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { }

    public void init(GLAutoDrawable gld) {
        GL gl = gld.getGL();
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        gl.glViewport(0, 0, 500, 300);
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glClearDepth(1.0);
        gl.glDepthFunc(GL.GL_LEQUAL);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);

        float mat_specular[] = { 0.5f, 0.5f, 0.5f, 0.5f };
        float mat_shininess[] = { 50.0f };
        gl.glShadeModel(GL.GL_SMOOTH);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, mat_specular, 0);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_specular, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, new float[] { 0.3f, 0.3f, 0.3f, 0.3f }, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_shininess, 0);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[] { 1.0f,    4.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[] { -10.0f, 10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT2, GL.GL_POSITION, new float[] { 10.0f, -10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT3, GL.GL_POSITION, new float[] { 10.0f,  10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT4, GL.GL_POSITION, new float[] { -10.0f, 10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT5, GL.GL_POSITION, new float[] { 10.0f, -10.0f,  -10.0f, 0.0f }, 0);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glEnable(GL.GL_LIGHT0);

        gl.glColorMaterial(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE);
        gl.glEnable(GL.GL_COLOR_MATERIAL);

        display(gld);
    }

    public abstract void _display(GLAutoDrawable drawable, GL gl);
    public final void display(GLAutoDrawable drawable) {
        GL gl = drawable.getGL();
        _display(drawable, gl);
        projection = Matrix.getProjectionMatrix(gl);
    }

    protected synchronized void updateVisibility(GL gl, Mesh mesh) {
        Matrix projection = Matrix.getProjectionMatrix(gl);
        IntBuffer buf = ByteBuffer.allocateDirect(9*4*4).order(ByteOrder.nativeOrder()).asIntBuffer();
        gl.glColor3f(0,0,0);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        for(Mesh.T t : mesh) t.glTriangle(gl, null);
        for(Mesh.Vertex v : mesh.vertices()) {
            Point p = v.getPoint();
            gl.glColor3f(1,1,1);
            gl.glBegin(gl.GL_POINTS);
            p.glVertex(gl);
            gl.glEnd();
            gl.glFlush();
            
            Point projected = projection.times(p);
            gl.glReadPixels((int)projected.x-1, (int)projected.y-1, 3, 3, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, buf);
            
            boolean vis = false;
            for(int j=0; j<9*4; j++) vis |= buf.get(j)!=0;
            v.visible = vis;
            if (vis) {
                gl.glColor3f(0,0,0);
                gl.glBegin(gl.GL_POINTS);
                p.glVertex(gl);
                gl.glEnd();
            }
        }
        gl.glShadeModel(GL.GL_SMOOTH);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glDrawBuffer(GL.GL_FRONT);
    }


    /** return the position of the mouse as a point in window-space */
    public Point getMouse() {
        return new Point(mousex, glcanvas.getHeight()-mousey, 0);
    }

    public void mouseWheelMoved(MouseWheelEvent e) {
        tz -= e.getWheelRotation();
    }

    public void keyTyped(KeyEvent e)  { }
    public void keyPressed(KeyEvent e)  {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = true; break;
            case KeyEvent.VK_ALT: alt = true; break;
            case KeyEvent.VK_SHIFT: shift = true; break;
        }
    }

    public void keyReleased(KeyEvent e) {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = false; break;
            case KeyEvent.VK_ALT: alt = false; break;
            case KeyEvent.VK_SHIFT: shift = false; break;
        }
    }

    public void mouseClicked(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mousePressed(MouseEvent e) {
        clickPoint = getMouse();
        clickClosest = closest == null ? null : closest.getPoint();
    }

    public void mouseReleased(MouseEvent e) {
        clickPoint = null;
        clickClosest = null;
    }

    public void mouseMoved(MouseEvent e) {
        mousex = e.getX();
        mousey = e.getY();
    }

    public void mouseDragged(MouseEvent e) {
        if (shift) {
            /*
            tx += (mousex - e.getX())/(float)20;
            ty += (mousey - e.getY())/(float)20;
            */
            if (closest != null && projection != null && clickClosest != null) {
                synchronized(this) {
                    Vec d1 = projection.inverse().times(getMouse()).minus(projection.inverse().times(clickPoint));
                    Vec delta = d1.plus(clickClosest).minus(closest.getPoint());
                    //System.out.println(delta + " " + closest.getPoint());
                    System.out.println(getMouse().minus(clickPoint));
                    closest.move(Matrix.translate(delta), false);
                }
            }
        } else {
            anglex -= mousex - e.getX();
            angley += mousey - e.getY();
        }
        mousex = e.getX();
        mousey = e.getY();
    }

    public Viewer(JFrame f) {
        this.f = f;
        glcanvas = new GLCanvas();
        glcanvas.addGLEventListener(this);
        f.add(glcanvas, BorderLayout.CENTER);
        glcanvas.addMouseListener(this);
        glcanvas.addMouseMotionListener(this);
        glcanvas.addMouseWheelListener(this);
        glcanvas.addKeyListener(this);
    }

    public void repaint() {
        glcanvas.repaint();
    }
}